/* 
 * --------- RNKC10 Servoansteuerung --------- 
 *  
 * Copyright (C) 2005 Fabian Greif <email@kreatives-chaos.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * --------- Beschreibung ---------
 *
 * Mit diesem Baustein lassen sich bis zu 10 Servos ber RS232 oder I2C 
 * Ansteuern. Die Auflsung betrgt dabei theoretisch ca. 4000 Schritte 
 * fr die Servos
 *
 * ----- Ansteuerung ber I2C ----- 
 * 
 * Der RNKC10 hat die Adresse 0x50 ( 80 ). Es mssten zustzlich zur Adresse
 * noch zwei weitere Parameter gesendet werden. Als erstes die 
 * Servonummer ( 0 bis 9 ), dann die Position des Servos ( 0 bis 255 )
 * 
 * Der RNKC10 untersttz bis jetzt leider noch nicht die normale 
 * I2C Geschwindigkeit ( 100 kHz ), die maximale Taktfrequnez betrgt im 
 * Moment ca. 20 kHz.
 *
 * Diese Einschrnkung resultiert aus der Servoansteuerung per Timerinterrupt.
 * Da dieser praktisch zu jedem Zeitpunkt im Hauptprogramm ausgelst werden kann
 * und gleich in die Interruptroutine gesprungen wird, kann es so passieren das
 * in dem I2C Slave Modul Signale verloren gehen.
 * Dieser Problem lsst sich erst mit einer Hardware basierten I2C Slave Lsung 
 * beheben ( geeigent sind zum Beispiel alle AVRs mit TWI oder USI Schnittstelle )
 * 
 * ----- Serielle Ansteuerung -----
 * 
 * Die Ansteuerung ber die serielle Schnittstelle erfolgt ganz hnlich zu der 
 * Ansteuerung per I2C. Zuerst werden zwei Syncronisationsbytes ( #S ) gesendet. 
 * Dann folgt die Nummer des Servos und dann die Position.
 * Es mssen also immer vier Bytes bertragen werden. Um zum Beispiel Servo 1 
 * in die Mittelstellung zu bringen mssten also folgende Werte gesendet werden:
 * 
 * 0x23 ( 35 )
 * 0x53 ( 83 )
 * 0x01 ( 1 )
 * 0x7F ( 127 )
 * 
 * Die Baudrate betrgt dabei 9600 Baud. Dies kann ganz einfach gendert werden wenn
 * im Programm den Wert fr UART_UBRR anpasst. Was fr diesen Wert einsetzten muss 
 * kann man am besten im Datenblatt unter "Uart Baudrate Examples" nachlesen.
 * 
 *
 * weitere Informationen findet man unter 
 *    http://www.kreatives-chaos.com/index.php?seite=rnkc10.
 */

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

// I2C Slave
#define	I2C_READ    	1
#define	I2C_WRITE   	0

#define	SDA				2
#define	SCL				3
#define	I2C_PORT		PORTD
#define	I2C_PIN			PIND
#define	I2C_DDR			DDRD

#define	I2C_DEV_ADRESSE	0x50

// Servoansteuerung
#define MAX_WERT		61536
#define	MIN_WERT		57536

// RS232 Empfang
#define	PREBYTE			'#'
#define	SYNCBYTE		'S'

#define	UART_UBRR		12		// Baudrate = 9600


void init(void);

void warte_auf_start_bedinung(void);
unsigned char empfange_adresse(void);
unsigned char empfange_daten(unsigned char erstes_bit);
void schreibe_daten(void);


volatile unsigned char
	servo_zaehler, nachricht_zaehler, nachricht_servo;
volatile unsigned int
	servo_werte[10];
	
int main(void)
{
	unsigned char 
		temp, temp2, i, schreiben, 
		wiederholter_start, adresse, 
		anzahl_daten, daten_array[2];
	
	init();
	
	while (1)
	{
		warte_auf_start_bedinung();
		
		// Empfangen der Adresse
		wiederholter_start = 1;
		while (wiederholter_start)
		{
			wiederholter_start = 0;
			anzahl_daten = 0;
			
			adresse = empfange_adresse();
			
			temp = adresse & 0xfe;
			if ( temp == I2C_DEV_ADRESSE )
			{
				// Nachricht ist fr diesen IC
				
				I2C_DDR |= (1<<SDA);		// Setze Acknowlegde
				// Warten bis SCL high
				while ((I2C_PIN & (1<<SCL)) != (1<<SCL));
				
				if ( ( adresse & 0x01 ) == I2C_WRITE )
				{
					schreiben = 1;
					while (schreiben)
					{
						// Warten bis SCL low
						while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
						I2C_DDR &= ~(1<<SDA);	// Lsche Acknowlegde
						
						// Warten bis SCL high
						while ((I2C_PIN & (1<<SCL)) != (1<<SCL));
						
						// erstes Bit empfangen
						i = I2C_PIN & (1<<SDA);
						
						// Abfrage ob es einen wiederholten Start gibt 
						// oder ob die Daten normal Empfangen werden knnen
						temp = I2C_PIN & ((1<<SDA)|(1<<SCL));
						temp2 = temp;
						while (temp2 == temp)
						{
							temp2 = I2C_PIN & ((1<<SDA)|(1<<SCL));
						}
						
						if ( !(temp2 & (1<<SCL) ) )
						{
							daten_array[anzahl_daten] = empfange_daten(i);
							anzahl_daten++;
						}
						else 
							schreiben = 0;
					}
					
					// Wiederholter Start
					if ( !(temp2 & (1<<SDA ) ) )
						wiederholter_start = 1;
				}
				else
				{
					// Warten bis SCL low
					while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
					I2C_DDR &= ~(1<<SDA);	// Lsche Acknowlegde
					
					schreibe_daten();
				}
			}
			
			// Auswertung
			if (anzahl_daten > 1)
			{
				if (daten_array[0] < 9)
					servo_werte[daten_array[0]] = daten_array[1] * ((MAX_WERT - MIN_WERT)/255) + MIN_WERT;
			}
		}
	}
	
	return 0;
}

void init(void)
{
	unsigned char i;
	
	DDRB = 0xff;
	PORTB = 0x00;
	
	DDRD = 0x72;
	PORTD = 0x00;
	
	// Timer1
	TCCR1A = 0x00;										// normaler Timermodus ( kein PWM )
	TCCR1B = 0x00;										// Timer1 deaktivieren
	TIMSK = (1<<TOIE1) | (1<<TOIE0);					// Timerberlaufinterrupts aktivieren
	
	// Timer0
	TCCR0 = (1<<CS02);									// Prescaler = 256
	TCNT0 = 0xde;
	
	// UART
	UCR = (1<<RXCIE) | (1<<RXEN);
	UBRR = UART_UBRR;
	
	// Servos in Mittelstellung bringen
	for (i=0;i<10;i++)
	{
		servo_werte[i] = ((MAX_WERT + MIN_WERT) / 2);
	}
	
	sei();												// Interrupts aktivieren
}

void warte_auf_start_bedinung(void)
{
	unsigned char temp, temp2 = 0;
	
	while ( temp2 != 0xaa )
	{
		// Warten bis SDA und SCL high
		temp = PIND & ((1<<SDA)|(1<<SCL));
		while (temp != ((1<<SDA)|(1<<SCL)) )
		{
			temp = PIND & ((1<<SDA)|(1<<SCL));
		}
		
		temp = PIND & ((1<<SDA)|(1<<SCL));
		
		// Warten bis sich etwas ndert
		temp2 = temp;
		while (temp2 == temp)
		{
			temp2 = PIND & ((1<<SDA)|(1<<SCL));
		}
		// Startbedinung ?
		if ( !(temp2 & (1<<SDA)) )
			temp2 = 0xaa;
	}
}

unsigned char empfange_adresse(void)
{
	unsigned char daten, i;
	
	daten = 0;
	
	for (i=0;i<8;i++)
	{
		daten <<= 1;
		
		// Warten bis SCL low
		while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
		
		// Warten bis SCL high
		while ((I2C_PIN & (1<<SCL)) != (1<<SCL));
		
		if (I2C_PIN & (1<<SDA))
			daten |= 1;
	}
	
	// Warten bis SCL low
	while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
	
	return daten;
}

unsigned char empfange_daten(unsigned char erstes_bit)
{
	unsigned char daten, i;
	
	daten = 0;
	if (erstes_bit)
		daten = 1;
	
	for (i=0;i<7;i++)
	{
		daten <<= 1;
		
		// Warten bis SCL low
		while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
		
		// Warten bis SCL high
		while ((I2C_PIN & (1<<SCL)) != (1<<SCL));
		
		if (I2C_PIN & (1<<SDA))
			daten |= 1;
	}
	
	// Warten bis SCL low
	while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
	
	I2C_DDR |= (1<<SDA);		// Setze Acknowlegde
	
	// Warten bis SCL high
	while ((I2C_PIN & (1<<SCL)) != (1<<SCL));
	
	// Warten bis SCL low
	while ((I2C_PIN & (1<<SCL)) == (1<<SCL));
	
	// Acknowlegde wird in der Hauptroutine gelscht
	
	return daten;
}

// FIXME Noch nicht implementiert
void schreibe_daten(void)
{
	
	
	
}

/*	Interrupt wird bei einem berlauf von Timer0 ausgelst ( alle 2,2 ms ) 
 *	Es werden nacheinander die 10 Servos ausgewhlt und angesteuert 
 *	Dauer : 17,5 S @ 4 MHz
 */
SIGNAL(SIG_OVERFLOW0) 
{
	unsigned char temp;
	
	temp = servo_zaehler;
	temp++;
	if (temp > 9)
		temp = 0;
	servo_zaehler = temp;
	
	switch(temp) 							// Signal fr die Servos auswhlen
	{
		case 0:
			PORTB |= (1<<PB0);
			break;
		case 1:
			PORTB |= (1<<PB1);
			break;
		case 2:
			PORTB |=  (1<<PB2);
			break;
		case 3:
			PORTB |=  (1<<PB3);
			break;
		case 4:
			PORTB |=  (1<<PB4);
			break;
		case 5:
			PORTB |=  (1<<PB5);
			break;
		case 6:
			PORTB |=  (1<<PB6);
			break;
		case 7:
			PORTB |=  (1<<PB7);
			break;
		case 8:
			PORTD |=  (1<<PD5);
			break;
		case 9:
			PORTD |=  (1<<PD6);
			break;
	}
	
	TCNT1 = servo_werte[temp];
	TCCR1B = (1<<CS10);							// Timer1 aktivieren, Prescaler = 1
	
	TCNT0 = 0xde;									// Timer0 neu laden
}

/*	Interrupt wird bei einem erlauf von Timer1 ausgelst
 *	Setzt das Signal fr die Servoansteuerung wieder zurck
 *	Dauer : 4,5 S @ 4 MHz
 *
 *	Timer1 : 
 *		61536  ->  1 ms
 *		59536  ->  1.5 ms
 *		57536  ->  2 ms
 */
SIGNAL(SIG_OVERFLOW1)
{

	PORTD &= ~((1<<PD5)|(1<<PD6));
	PORTB = 0x00;									// Servosignal wieder lschen
	
	TCCR1B = 0x00;									// Timer1 deaktivieren
}

// 	Interrupt wird ausgelst wenn ein Byte empfangen wurde
SIGNAL(SIG_UART_RECV)
{
	unsigned char buffer;
	
	buffer = UDR;
	
	switch (nachricht_zaehler)
	{
		case 0:
			if (buffer == PREBYTE)
				nachricht_zaehler = 1;
			break;
		case 1:
			if (buffer == SYNCBYTE)
				nachricht_zaehler = 2;
			break;
		case 2:
			if (nachricht_servo > 9)
			{
				nachricht_zaehler = 0;
			}
			else
			{
				nachricht_servo = buffer;
				nachricht_zaehler = 3;
			}
			break;
		case 3:
			servo_werte[nachricht_servo] = buffer * ((MAX_WERT - MIN_WERT)/255) + MIN_WERT;
			nachricht_zaehler = 0;
			break;
		default:
			nachricht_zaehler = 0;
			break;
	}
}
